#include "4pcs.h"

#include <fstream>
#include <iostream>
#include <string>
#include <opencv2/highgui/highgui.hpp>

#define sqr(x) ((x) * (x))

using namespace std;
using namespace match_4pcs;

// First input.
std::string input1 = "input1.obj";

// Second input.
std::string input2 = "input2.obj";

// Output. The transformed second input.
std::string output = "output.obj";

// Delta (see the paper).
double delta = 5.0;

// Estimated overlap (see the paper).
double overlap = 0.2;

// Threshold of the computed overlap for termination. 1.0 means don't terminate
// before the end.
double thr = 1.0;

// Maximum norm of RGB values between corresponded points. 1e9 means don't use.
double max_color = 150;

// Number of sampled points in both files. The 4PCS allows a very aggressive
// sampling.
int n_points = 200;

// Maximum angle (degrees) between corresponded normals.
double norm_diff = 90.0;

// Maximum allowed computation time.
int max_time_seconds = 10;

// For IO. We want to keep the 4pcs itself free from IO.
struct tripple {
  int a;
  int b;
  int c;
  int n1;
  int n2;
  int n3;
  int t1;
  int t2;
  int t3;
  tripple() {}
  tripple(int _a, int _b, int _c) : a(_a), b(_b), c(_c) {}
};

void getArgs(int argc, char **argv) {
  int i = 1;
  while (i < argc) {
    if (!strcmp(argv[i], "-i")) {
      input1 = std::string(argv[++i]);
      input2 = std::string(argv[++i]);
    } else if (!strcmp(argv[i], "-o")) {
      overlap = atof(argv[++i]);
    } else if (!strcmp(argv[i], "-d")) {
      delta = atof(argv[++i]);
    } else if (!strcmp(argv[i], "-c")) {
      max_color = atof(argv[++i]);
    } else if (!strcmp(argv[i], "-t")) {
      max_time_seconds = atoi(argv[++i]);
    } else if (!strcmp(argv[i], "-a")) {
      norm_diff = atof(argv[++i]);
    } else if (!strcmp(argv[i], "-n")) {
      n_points = atoi(argv[++i]);
    } else if (!strcmp(argv[i], "-h")) {
      fprintf(stderr, "\nUsage: %s -i input1 input2\n\n", argv[0]);
      fprintf(stderr, "        [ -o overlap (%2.2f) ]\n", overlap);
      fprintf(stderr, "        [ -d delta (%2.2f) ]\n", delta);
      fprintf(stderr, "        [ -n n_points (%d) ]\n", n_points);
      fprintf(stderr, "        [ -a norm_diff (%f) ]\n", norm_diff);
      fprintf(stderr, "        [ -c max_color_diff (%f) ]\n", max_color);
      fprintf(stderr, "        [ -t max_time_seconds (%d) ]\n",
              max_time_seconds);
      exit(0);
    } else if (argv[i][0] == '-') {
      cerr << "Unknown flag\n";
      exit(-1);
    };
    i++;
  }
}

// Obj read/write simple functions.
bool ReadObject(char *name, vector<Point3D> &v, vector<cv::Point2f> &tex_coords,
                vector<cv::Point3f> &normals, vector<tripple> &tris,
                vector<std::string> &mtls) {
  fstream f(name, ios::in);
  if (!f || f.fail()) return false;
  char str[1024];
  float x, y, z;
  v.clear();
  tris.clear();
  while (!f.eof()) {
    f.getline(str, 1023);
    char ch[128];
    sscanf(str, "%s %*s", ch);
    if (strcmp(ch, "v") == 0) {
      sscanf(str, "%s %f %f %f", ch, &x, &y, &z);
      v.push_back(Point3D(x, y, z));
      v[v.size() - 1].set_rgb(cv::Vec3f(0, 0, 0));
    } else if (strcmp(ch, "vt") == 0) {
      cv::Point2f tex_coord;
      sscanf(str, "%s %f %f", ch, &tex_coord.x, &tex_coord.y);
      tex_coords.push_back(tex_coord);
    } else if (strcmp(ch, "vn") == 0) {
      cv::Point3f normal;
      sscanf(str, "%s %f %f %f", ch, &(normal.x), &(normal.y), &(normal.z));
      normals.push_back(normal);
    } else if (strcmp(ch, "f") == 0) {
      tripple triangle;
      if (normals.size() && !tex_coords.size()) {
        sscanf(str, "%s %d//%d %d//%d %d//%d", ch, &(triangle.a),
               &(triangle.n1), &(triangle.b), &(triangle.n2), &(triangle.c),
               &(triangle.n3));
      } else if (normals.size() && tex_coords.size()) {
        sscanf(str, "%s %d/%d/%d %d/%d/%d %d/%d/%d", ch, &(triangle.a),
               &(triangle.t1), &(triangle.n1), &(triangle.b), &(triangle.t2),
               &(triangle.n2), &(triangle.c), &(triangle.t3), &(triangle.n3));
      } else if (!normals.size() && tex_coords.size()) {
        sscanf(str, "%s %d/%d %d/%d %d/%d", ch, &(triangle.a), &(triangle.t1),
               &(triangle.b), &(triangle.t2), &(triangle.c), &(triangle.t3));
      } else if (!normals.size() && !tex_coords.size()) {
        sscanf(str, "%s %d %d %d", ch, &(triangle.a), &(triangle.b),
               &(triangle.c));
      }
      tris.push_back(triangle);
      if (normals.size()) {
        v[triangle.a - 1].set_normal(normals[triangle.n1 - 1]);
        v[triangle.b - 1].set_normal(normals[triangle.n2 - 1]);
        v[triangle.c - 1].set_normal(normals[triangle.n3 - 1]);
      }
    } else if (strcmp(ch, "mtllib") == 0) {
      mtls.push_back(str + 7);
    }
  }
  f.close();

  if (mtls.size()) {
    f.open(mtls[0].c_str(), ios::in);
    while (f && !f.fail()) {
      std::string img_name, dummy;
      f >> dummy;
      if (strcmp(dummy.c_str(), "map_Kd") == 0) {
        f >> img_name;
        cv::Mat tex = cv::imread(img_name);
        if (!tex.empty()) {
          for (int i = 0; i < tris.size(); ++i) {
            const tripple &t = tris[i];
            cv::Point2f tc1 = tex_coords[t.t1 - 1];
            cv::Point2f tc2 = tex_coords[t.t2 - 1];
            cv::Point2f tc3 = tex_coords[t.t3 - 1];
            if (tc1.x < 1.0 && tc1.x > 0 && tc1.y < 1.0 && tc1.y > 0 &&
                tc2.x < 1.0 && tc2.x > 0 && tc2.y < 1.0 && tc2.y > 0 &&
                tc3.x < 1.0 && tc3.x > 0 && tc3.y < 1.0 && tc3.y > 0) {
              v[t.a - 1].set_rgb(cv::Vec3f(
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc1.y * tex.rows, tc1.x * tex.cols)[0]),
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc1.y * tex.rows, tc1.x * tex.cols)[1]),
                  static_cast<float>(tex.at<cv::Vec3b>(tc1.y * tex.rows,
                                                       tc1.x * tex.cols)[2])));

              v[t.b - 1].set_rgb(cv::Vec3f(
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc2.y * tex.rows, tc2.x * tex.cols)[0]),
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc2.y * tex.rows, tc2.x * tex.cols)[1]),
                  static_cast<float>(tex.at<cv::Vec3b>(tc2.y * tex.rows,
                                                       tc2.x * tex.cols)[2])));

              v[t.c - 1].set_rgb(cv::Vec3f(
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc3.y * tex.rows, tc3.x * tex.cols)[0]),
                  static_cast<float>(
                      tex.at<cv::Vec3b>(tc3.y * tex.rows, tc3.x * tex.cols)[1]),
                  static_cast<float>(tex.at<cv::Vec3b>(tc3.y * tex.rows,
                                                       tc3.x * tex.cols)[2])));
            }
          }
        }
      }
    }
  }
  f.close();
  if (v.size() == 0) return false;
  return true;
}

bool WriteObject(char *name, vector<Point3D> &v,
                 vector<cv::Point2f> &tex_coords, vector<cv::Point3f> &normals,
                 vector<tripple> &tris, vector<std::string> &mtls) {
  fstream f(name, ios::out);
  if (!f || f.fail()) return false;
  int i;

  normals.clear();

  for (i = 0; i < mtls.size(); ++i) {
    f << "mtllib " << mtls[i] << endl;
  }

  for (i = 0; i < v.size(); ++i) {
    f << "v " << v[i].x << " " << v[i].y << " " << v[i].z << endl;
  }

  for (i = 0; i < normals.size(); ++i) {
    f << "vn " << normals[i].x << " " << normals[i].y << " " << normals[i].z
      << endl;
  }

  for (i = 0; i < tex_coords.size(); ++i) {
    f << "vt " << tex_coords[i].x << " " << tex_coords[i].y << endl;
  }

  for (i = 0; i < tris.size(); ++i) {
    if (!normals.size() && !tex_coords.size())
      f << "f " << tris[i].a << " " << tris[i].b << " " << tris[i].c << endl;
    else if (tex_coords.size())
      f << "f " << tris[i].a << "/" << tris[i].t1 << " " << tris[i].b << "/"
        << tris[i].t2 << " " << tris[i].c << "/" << tris[i].t3 << endl;
  }

  f.close();
  return true;
}

int main(int argc, char **argv) {
  vector<Point3D> set1, set2;
  vector<cv::Point2f> tex_coords1, tex_coords2;
  vector<cv::Point3f> normals1, normals2;
  vector<tripple> tris1, tris2;
  vector<std::string> mtls1, mtls2;

  getArgs(argc, argv);

  // Read the inputs.
  if (!ReadObject((char *)input1.c_str(), set1, tex_coords1, normals1, tris1,
                  mtls1)) {
    perror("Can't read input set1");
    exit(-1);
  }

  if (!ReadObject((char *)input2.c_str(), set2, tex_coords2, normals2, tris1,
                  mtls2)) {
    perror("Can't read input set2");
    exit(-1);
  }

  // Our matcher.
  Match4PCSOptions options;

  // Set parameters.
  cv::Mat mat = cv::Mat::eye(4, 4, CV_64F);
  options.overlap_estimation = overlap;
  options.sample_size = n_points;
  options.max_normal_difference = norm_diff;
  options.max_color_distance = max_color;
  options.max_time_seconds = max_time_seconds;
  Match4PCS matcher(options);

  // Match and return the score (estimated overlap or the LCP).
  float score = matcher.ComputeTransformation(set1, &set2, &mat);

  printf("Score: %f\n", score);
  printf("Merged object was written to %s\n", output.c_str());
  printf("(Homogeneous) Transformation from %s to %s:\n", input2.c_str(),
         input1.c_str());
  printf(
      "\n\n%25.3f %25.3f %25.3f %25.3f\n%25.3f %25.3f %25.3f %25.3f\n%25.3f "
      "%25.3f %25.3f %25.3f\n%25.3f %25.3f %25.3f %25.3f\n\n",
      mat.at<double>(0, 0), mat.at<double>(0, 1), mat.at<double>(0, 2),
      mat.at<double>(0, 3), mat.at<double>(1, 0), mat.at<double>(1, 1),
      mat.at<double>(1, 2), mat.at<double>(1, 3), mat.at<double>(2, 0),
      mat.at<double>(2, 1), mat.at<double>(2, 2), mat.at<double>(2, 3),
      mat.at<double>(3, 0), mat.at<double>(3, 1), mat.at<double>(3, 2),
      mat.at<double>(3, 3));

  // If the default images are the input then we need to test the result.
  if (input1 == "input1.obj" && input2 == "input2.obj") {  // test!
    float gt_mat[16] = {0.977,  -0.180,  -0.114, 91.641, 0.070, 0.778,
                        -0.624, 410.029, 0.201,  0.602,  0.773, 110.810,
                        0.000,  0.000,   0.000,  1.000};
    float norm_val = 0.0;
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        norm_val += sqr(mat.at<double>(i, j) - gt_mat[i * 4 + j]);
      }
    }
    assert(norm_val > 9.46881e-10);
  }

  // Write the transformed input2 as the output.
  WriteObject((char *)output.c_str(), set2, tex_coords2, normals2, tris1,
              mtls2);

  return 0;
}
